/*
 *
 * gobj_i.h
 *
 */


#ifndef GOBJ_I_H
#define GOBJ_I_H

#include <iostream.h>
#include <math.h>

// ----------------------------------------------------------------
//
// MVector
//
// ----------------------------------------------------------------

inline MVector::MVector(int xx, int yy, int zz)
{
  x = xx;
  y = yy;
  z = zz;
}


inline void MVector::Normalize(void)
{
  double xx = x; 
  double yy = y; 
  double zz = z; 

  double k = sqrt(xx*xx + yy*yy + zz*zz);
  if(k)
    {
      k = (double)M_BIG / k;
      x = (int)(k * xx);
      y = (int)(k * yy);
      z = (int)(k * zz);
    }
}


inline void MVector::RotX(Angle r)
{
   long t1 = y;
   long t2 = z;
   long s = Sin(r),c = Cos(r);

   y = ( c * t1 + s * t2) >> SIN_SHIFT;
   z = (-s * t1 + c * t2) >> SIN_SHIFT;
}

inline void MVector::RotY(Angle r)
{
  long t1 = x;
  long t2 = z;
  long s = Sin(r),c = Cos(r);

  x = ( c * t1 + s * t2) >> SIN_SHIFT;
  z = (-s * t1 + c * t2) >> SIN_SHIFT;
 }

inline void MVector::RotZ(Angle r)
{
   long t1 = x;
   long t2 = y;
   long s = Sin(r),c = Cos(r);

   x = ( c * t1 + s * t2) >> SIN_SHIFT;
   y = (-s * t1 + c * t2) >> SIN_SHIFT;
}

inline void MVector::Rotate(Angle rx, Angle ry, Angle rz)
{
  RotX(rx);
  RotY(ry);
  RotZ(rz);
}

inline void MVector::Translate(int tx, int ty, int tz)
{
  x += tx;
  y += ty;
  z += tz;
}

inline void MVector::Scale(double sx, double sy, double sz)
{
  x = (int)(sx *(double)x);  
  y = (int)(sy *(double)y);
  z = (int)(sz *(double)z);
}

inline void MVector::Scale(double s)
{
  Scale(s,s,s);
}


inline istream& operator >> (istream& input,  MVector& V)
{
  double x,y,z;
  input >> x >> y >> z;
  x = (int)(x);
  y = (int)(y);
  z = (int)(z);
}

inline ostream& operator << (ostream& output, MVector& V)
{
  double x = V.x, y = V.y, z = V.z;
  output << x << ' ' << y << ' ' << z << endl;
}


// ----------------------------------------------------------------
//
// MVertex
//
// ----------------------------------------------------------------


inline MVertex::MVertex(int xx, int yy, int zz) : MVector(xx,yy,zz)
{
}

// ----------------------------------------------------------------
//
// GraphicObject
//
// ----------------------------------------------------------------

inline void GraphicObject::Rotate(char axis, Angle r)
{
  switch(axis)
    {
    case 'x':
    case 'X':
      RotX(r);
      break;
    case 'y':
    case 'Y':
      RotY(r);
      break;
    case 'z':
    case 'Z':
      RotZ(r);
      break;
    default:
      cerr << "GraphicObject::Rotate : invalid axis" << endl;
      break;
    }
}
 

inline void GraphicObject::Rotate(Angle rx, Angle ry, Angle rz)
{
  RotX(rx);
  RotY(ry);
  RotZ(rz);
}
 
inline void GraphicObject::Scale(double s)
{
  Scale(s,s,s);
}


inline PolygonEngine* operator << (PolygonEngine* output, GraphicObject& O)
{
  O.Draw(output);
  return output;
}

// ----------------------------------------------------------------
//
// Shading parameters (will move later on ...)
//
// ----------------------------------------------------------------


inline void GraphicObject::SetGeometry(int width, int height)
{
  _width  = width;
  _height = height;
}

inline int Clamp(int v, int m, int M)
{
  return (v < m) ? m : ((v > M) ? M : v);
}

inline void GraphicObject::SetZoom(int  zoom)
{
  _zoom = Clamp(zoom, 1, 100);
}

inline int  GraphicObject::GetZoom(void)
{
  return _zoom;
}

inline void GraphicObject::SetZoomStep(int zoom_step)
{
  _zoom_step = Clamp(zoom_step, 1, 10);
}

inline int GraphicObject::GetZoomStep(void)
{
  return _zoom_step;
}

inline void GraphicObject::ZoomIn(void)
{
  _zoom += _zoom_step;
  _zoom = Clamp(_zoom, 1, 100);
}

inline void GraphicObject::ZoomOut(void)
{
  _zoom -= _zoom_step;
  _zoom = Clamp(_zoom, 1, 100);
}

inline void GraphicObject::SetLightDirection(int  Lx, int  Ly, int  Lz)
{
  _L = MVector(Lx, Ly, Lz);
}

inline void GraphicObject::GetLightDirection(int *Lx, int *Ly, int *Lz)
{
  *Lx = _L.x;
  *Ly = _L.y;
  *Lz = _L.z;
}

inline void GraphicObject::SetLightColor(int  Lr, int  Lg, int  Lb)
{
  _Lr = Lr;
  _Lg = Lg;
  _Lb = Lb;
}

inline void GraphicObject::GetLightColor(int *Lr, int *Lg, int *Lb)
{
  *Lr = _Lr;
  *Lg = _Lg;
  *Lb = _Lb;
}

inline void GraphicObject::SetLightColor(int  Lc)
{
  _Lc = Lc;
}

inline void GraphicObject::GetLightColor(int *Lc)
{
  *Lc = _Lc;
}

inline void GraphicObject::Project(MVertex *V)
{
  const long YE = 65536 >> 9;
  
  V->Projection.x = (_width  >> 1) + 
                 (((((V->x*YE*3) / ((1 << 23)/_width ))) * _zoom) >> 4);
  V->Projection.y = (_height >> 1) + 
                 (((((V->y*YE*3) / ((1 << 23)/_width ))) * _zoom) >> 4);
  V->Projection.z = (M_BIG - V->z ) >> 5;
}


inline MVector& GraphicObject::LightDirection(void)
{
  return _L;
}

#endif
