
/*
 *
 * rotate.C
 *
 * This program demonstrates TAGL polygon drawing and 
 * shading abilities. 
 *
 */


#include <fstream.h>
#include <stdlib.h>
#include "gport.h"
#include "polyeng.h"
#include "mesh.h"
#include <time.h>

GraphicPort   *GP;
PolygonEngine *PE;
Mesh m;


int do_lighting = 1;

inline void ColorToggle(Mesh &m)
{
  static int color = 0;

  int i;

  if(color)
    {
      GP->ColormapMode();
      PE->ColormapMode();

      for(i=0; i<64; i++)
	GP->MapColor(i, i << 2, i << 2, i << 2);

      m.Mode().Reset(MF_COLOR);
      color = 0;
    }
  else
    {
      GP->RGBMode();
      PE->RGBMode();
      m.Mode().Set(MF_COLOR);
      color = 1;
    }
}

inline void Toggle(Mesh &mesh, Flag f)
{
  if(mesh.Mode().Get(f))
    mesh.Mode().Reset(f);
  else
    mesh.Mode().Set(f);
}

inline void Toggle(PolygonEngine *pe, Flag f)
{
  if(pe->Attributes().Get(f))
    pe->Attributes().Reset(f);
  else
    pe->Attributes().Set(f);
  pe->CommitAttributes();
}

inline void Toggle(int &flag)
{
  flag = flag ? 0 : 1;
}

void DoCommand(char q)
{
  switch(q)
    {
    case '+':
      m.ZoomIn();
      break;
    case '-':
      m.ZoomOut();
      break;
    case 'w':
      Toggle(m, MF_SMOOTH);
      break;
    case 'n':
      m.InvertNormals();
      break;
    case 'd':
      Toggle(PE,GA_DITHER);
      break;
    case 'z':
      Toggle(m, MF_CONVEX);
      break;
    case 'c':
      Toggle(m, MF_CLOSED);
      break;
    case 'a':
      Toggle(m, MF_WIREFRAME);
      break;
    case 'l':
      Toggle(do_lighting);
      break;
    case ' ':
      ColorToggle(m);
      break;
    default:
      break;
    }
}

int main(int argc, char *argv[])
{

  int i;
  int elapsed = 0;
  int frames  = 0;

  GP = MakeGraphicPort("TAGL 3D object viewer", 320, 200);
  PE = MakePolygonEngine(GP);
  
  if(!GP)
    {
      cerr << "cannot open graphic port\n";
      exit(-1);
    }

  if(!PE)
    {
      cerr << "cannot create polygon engine\n";
      exit(-1);
    }

  PE->ColormapMode();

  for(i=0; i<64; i++)
    GP->MapColor(i, i << 2, i << 2, i << 2);

  GP->ZBuffer(1);

  if(argc > 3)
    {
      cerr << "wrong number of arguments" << endl;
      cerr << "rotate geometry.geom [colors.ipcol]" << endl;
      exit(-1);
    }

  ifstream input(argv[1]);

  if(!input)
    {
      cerr << "fatal: cannot open geometry file" << endl;
      exit(-1);
    }

  input >> m;

  if(m.ErrorCode())
    {
      cerr << "geometry file: error code #" << m.ErrorCode() << endl;
      exit(-1);
    }

  if(argc > 2)
    {
      ifstream input(argv[2]);
      if(!input)
	  cerr << "cannot open color file" << endl;
      else
	{
	  input >> m;
	  if(m.ErrorCode())
	    {
	      cerr << "color file: error code #" << m.ErrorCode() << endl;
	      exit(-1);
	    }
	  else
	    {
	      cerr << "blending object colors ..." << endl;
	      m.Blend();
	      cerr << "blending completed." << endl;
	    }
	}
    }

  cerr << "smoothing object ..." << endl;
  m.Smooth();
  m.Mode().Set(MF_CLOSED);
  cerr << "object processed, ready to render." << endl;

  elapsed = time(NULL);

  char q = 0;

  int oldx = 0, oldy = 0;
  int btn;
  int dx, dy;
  int translating    = 0;
  int rotating       = 0;
  int rotating_light = 0;

  while(q != 'q')
    {
      int mx, my;
      dx = mx - oldx;
      dy = my - oldy;
      oldx = mx;
      oldy = my;


//      GP->WaitEvent();
      q   = GP->GetKey(); 
      btn = GP->GetMouse(&mx, &my);

      DoCommand(q);

      if(btn & 1)
	{
	  if(rotating)
	    m.Rotate(dy >> 2,dx >> 2,0);
	  else
	    rotating = 1;
	}
      else
	rotating = 0;
      
      if(btn & 2)
	{
	  if(translating)
	    m.Translate(dx << 4, dy << 4, 0);
	  else
	    translating = 1;
	}
      else
	translating = 0;

      if(btn & 4)
	{
	  if(rotating_light)
	    m.LightDirection().Rotate((dy) >> 2,(dx) >> 2,0);
	  else
	    rotating_light = 1;
	}
      else
	rotating_light = 0;
  
      if(do_lighting)
	m.Lighting();

      m.Setup(PE);
      
      GP->Clip().Set(10,10,GP->Width() - 10, GP->Height() - 10);
      PE << m;
      GP->SwapBuffers();
      frames++;
    }

  elapsed = time(NULL) - elapsed;

  cerr << " elapsed: " << elapsed << " frames: " << frames 
       << " fps: " << (double)frames/(double)elapsed << endl; 

  delete PE;
  delete GP;

}

