
/*
 *
 * interface between tagl and wt.
 * allows both X11 and svga graphics.
 *
 */


extern "C" {
#include "wt.h"
#include "error.h"
#include "wtmem.h"
#include "framebuf.h"
#include "graphics.h"
#include "dither.h"
}

#include <gport.h>
#include <stdlib.h>

static Palette_info load_palette(void);
static int graphics_initialized = 0;
static Boolean dither_mode = False, 
               blocky_mode = False, 
               direct_mode = False;
static Pixel       *render_fb;
static DevicePixel *display_fb;
static int display_height, display_width;
static Boolean truecolor_mode = False;

GraphicPort *GP;


Graphics_info *init_graphics(int width, int height, Boolean blocky)
{
     static Graphics_info ginfo;

     if (graphics_initialized == 1)
	  return &ginfo;

     ginfo.width = width;
     ginfo.height = height;

     blocky_mode = blocky;

     if (blocky_mode) 
       {
	 display_width = ginfo.width * 2;
	 display_height = ginfo.height * 2;
       } 
     else 
       {
	 display_width = ginfo.width;
	 display_height = ginfo.height;
       }


     GP = MakeGraphicPort("What is that ?", 
			  display_width, display_height, MSG_ALL);

     if(GP->BitsPerPixel() > 16)
       {
	 cerr << "Sorry, display depth not handled (yet), exiting ...\n";
	 delete GP;
	 exit(-1);
       }


// I HATE those boolean types !!!!

     truecolor_mode = GP->BitsPerPixel() > 8 ? True : False;
//     direct_mode = GP->SingleBuffer() ? True : False;

     direct_mode = False;

     if (!truecolor_mode && sizeof(Pixel) != sizeof(DevicePixel))
	 dither_mode = True;


     display_width  = GP->Width();
     display_height = GP->Height(); 

     if(blocky_mode)
       {
	 ginfo.width = GP->Width() >> 1;
	 ginfo.height = GP->Height() >> 1;
       }
     else
       {
	 ginfo.width = GP->Width();
	 ginfo.height = GP->Height();
       }

     ginfo.palette = load_palette();

     display_fb = (DevicePixel *)GP->GraphMem(); 

     if(dither_mode || blocky_mode)
       render_fb =  (Pixel *)wtmalloc(width * height * sizeof(Pixel));

     if(truecolor_mode)
       render_fb = (Pixel *)display_fb;

     if (dither_mode)
	  calc_dither(&ginfo);

     graphics_initialized = 1;

     return &ginfo;
}


static Palette_info load_palette(void)
{
     int color_levels = 6;
     int r, g, b;
     int index = 0;
     Palette_info pinfo;


     pinfo.rgb_cube_size = color_levels;
     pinfo.color_lookup = (long *)wtmalloc(sizeof(int) * color_levels *
				           color_levels * color_levels);

     for (r = 0; r < color_levels; r++)
       for (g = 0; g < color_levels; g++)
	 for (b = 0; b < color_levels; b++) 
	   {
	     int red, green, blue;

	     red =   (short) (r * 255 / (color_levels - 1));
	     green = (short) (g * 255 / (color_levels - 1));
	     blue =  (short) (b * 255 / (color_levels - 1));

	     GP->MapColor(index, red, green, blue);

	     pinfo.color_lookup[index] = GP->Colormap()[index];
	     index++;
     }

     return pinfo;
}

     
void end_graphics(void)
{
  if (graphics_initialized) 
    {
      delete GP;
      graphics_initialized = 0;
    }
}

void update_screen(Framebuffer *fb)
{

  if (blocky_mode) 
    {
      if (dither_mode)
	expand_dither(render_fb, display_fb,
		      fb->fb_width, fb->fb_height);
      else
	expand(render_fb, display_fb, fb->fb_width, fb->fb_height);
    } 
  else 
    {
      if (dither_mode)
	dither(render_fb, display_fb, fb->fb_width, fb->fb_height);
    }

  if(!direct_mode)
    GP->SwapBuffers();
}


Pixel *get_framebuffer_memory(int width, int height)
{
  return (dither_mode || blocky_mode) ? 
         (Pixel *)render_fb :
         (Pixel *)display_fb;
}
