/*
 *
 * SVGAgport.C
 *
 */


#include "SVGAgport.h"
#include <stdlib.h>
#include <vgamouse.h>

SVGAGraphicPort::SVGAGraphicPort(char *name, 
				 ScrCoord width, ScrCoord height, int verbose_level): 
       GraphicPort(name, width, height, verbose_level)
{

  int i;

  _proc_name = "SVGAGraphicPort";

  _error_code = SGPE_OK;

  _width  = width;
  _height = height;

  _bits_per_pixel  = 8;
  _bytes_per_pixel = 1;

  vga_init();
  vga_setmousesupport(1);

  if(!SetupScreen())
    return;

  if(!AllocateImgBuffer())
    return;

  Attributes().Set(GPA_DBUFF);

  ColormapMode();

  _clip.Set(0,0,_width-1,_height-1);

  _graph_mem = _v_graph_mem;

  for(i=0; i<256; i++)
    _colormap[i] = i;
}


////
////
//
//   Attributes handling & destructor
//
////
////

SVGAGraphicPort::~SVGAGraphicPort(void)
{
  if(_resources.Get(SGPR_IMGBUFFER))
    FreeImgBuffer();

  if(_resources.Get(SGPR_ZBUFFER))
    FreeZBuffer();

  vga_setmode(TEXT);
}

void SVGAGraphicPort::CommitAttributes(void)
{
  Flags changed;

  changed.SetAll(_last_attributes.GetAll() ^ Attributes().GetAll());

  if(changed.Get(GPA_CMAP) && Attributes().Get(GPA_CMAP))
    ColormapMode();

  if(changed.Get(GPA_RGB)  && Attributes().Get(GPA_RGB))
    RGBMode();

  if(changed.Get(GPA_ZBUFF))
    ZBuffer(Attributes().Get(GPA_ZBUFF));

  if(changed.Get(GPA_DBUFF))
    {
      if(Attributes().Get(GPA_DBUFF))
	DoubleBuffer();
      else
	SingleBuffer();
    }
}

////
////
//
//   Set video mode 
//
////
////


int SVGAGraphicPort::SetupScreen() 
{
  
  if((_mode = vga_getdefaultmode()) < 0)
    if(_width <= 320)
      _mode = G320x200x256;
    else
      if(_width <= 640)
	_mode = G640x480x256;
      else
	if(_width <= 800)
	  _mode = G800x600x256;
	else
	  _mode = G1024x768x256;

  if(!vga_hasmode(_mode))
    {
      (*this)[MSG_WARNING] << "could not open " << _width << "*" << _height << " screen\n";
      (*this)[MSG_WARNING] << "trying 320x200 ...\n";

      _mode = G320x200x256;

      if(!vga_hasmode(_mode))
	{
	  (*this)[MSG_ERROR] << "seems you've got an antique graphic board...\n";
	  _error_code = SGPE_MODE;
	  return 0;
	}
    }

  vga_setmode(_mode);
  _modeinfo = vga_getmodeinfo(_mode);


  _linear = _modeinfo->flags & (1 << 4);
  if(_linear)
    vga_setlinearaddressing();
    
  _linear = _linear || (_mode == G320x200x256);

  if(_linear)
    (*this)[MSG_INFO] << "Linear adressing mode is available. Good.\n";

  _width           = _modeinfo->width;
  _height          = _modeinfo->height;
  _bytes_per_line  = _modeinfo->linewidth;
  _bytes_per_pixel = _modeinfo->bytesperpixel;

  _size  = _width * _height * _bytes_per_pixel;
  _pages = _size >> 16;
  _pad   = _size & 65535;

  _r_graph_mem = vga_getgraphmem();

  (*this)[MSG_INFO] << "width=" << _width << " height=" << _height << " bpl=" 
                    << _bytes_per_line << "\n";


  switch(_bytes_per_pixel)
    {
    case 1:
      (*this)[MSG_INFO] << "8bpp mode\n";
      _R_mask = 0;
      _G_mask = 0;
      _B_mask = 0;
      _bits_per_pixel = 8;
      break;
    case 2:
      (*this)[MSG_INFO] << "16bpp HiColor mode, waoow\n";
      (*this)[MSG_WARNING] << "Assuming weight is 555\n";
      _bits_per_pixel = 15;
      _R_mask = 31 << 10;
      _G_mask = 31 << 5;
      _B_mask = 31;
      break;
    case 3:
    case 4:
      (*this)[MSG_INFO] << "24-32bpp TrueColor mode. You've got the high score\n";
      _bits_per_pixel = 24;
      _R_mask = 255 << 16;
      _G_mask = 255 << 8;
      _B_mask = 255;
      break;
    }

  return 1;
}

////
////
//
// ImgBuffer utilities
//
////
////

int SVGAGraphicPort::AllocateImgBuffer(void)
{
  if(_resources.Get(SGPR_IMGBUFFER))
    {
      (*this)[MSG_WARNING] << "Img buffer already allocated\n";
      return 1;
    }

  (*this)[MSG_RESOURCE] << "allocating image buffer\n";
  _v_graph_mem =  new ColorIndex[_bytes_per_line * _height];
  if(!_v_graph_mem)
    {
      (*this)[MSG_WARNING] << "could not alloc image buffer\n";
      _error_code = SGPE_MALLOC;
      return 0;
    }
  _resources.Set(SGPR_IMGBUFFER);
  return 1;
}

void SVGAGraphicPort::FreeImgBuffer(void)
{
  if(!_resources.Get(SGPR_IMGBUFFER))
    {
      (*this)[MSG_WARNING] << "Img buffer already freed or not allocated\n";
      return;
    }
  (*this)[MSG_RESOURCE] << "freeing image buffer\n";
  delete[] _v_graph_mem;
  _resources.Reset(SGPR_IMGBUFFER);
}


////
////
//
// ZBuffer 
//
////
////

int SVGAGraphicPort::AllocateZBuffer(void)
{
  if(_resources.Get(SGPR_ZBUFFER))
    {
      (*this)[MSG_WARNING] << "ZBuffer already allocated\n";
      return 1;
    }

  if(!(_z_mem = new ZCoord[_bytes_per_line*_height]))
    {
      (*this)[MSG_ERROR] << "could not alloc ZBuffer\n";
      _error_code = SGPE_MALLOC;
      return 0;
    }

  (*this)[MSG_RESOURCE] << "ZBuffer allocated\n";
  _resources.Set(SGPR_ZBUFFER);
  return 1;
}

void SVGAGraphicPort::FreeZBuffer(void)
{
  if(!_resources.Get(SGPR_ZBUFFER))
    {
      (*this)[MSG_WARNING] << "ZBuffer already freed or not allocated\n";
      return;
    }
  (*this)[MSG_RESOURCE] << "Freeing ZBuffer\n";
  delete[] _z_mem;
  _resources.Reset(SGPR_ZBUFFER);
}


////
////
//
// Resize frame buffer & Z buffer
//
////
////


int SVGAGraphicPort::Resize(const ScrCoord width, const ScrCoord height)
{
  int retval = 1;

  (*this)[MSG_INFO] << "Resizing to " << width << "x" << height << "\n";

  _width  = width;
  _height = height;
  _bytes_per_line = width;

  SetupScreen();

  if(Attributes().Get(GPA_DBUFF))
    {
      FreeImgBuffer();
      retval = AllocateImgBuffer();
    }

  if(Attributes().Get(GPA_ZBUFF))
    {
      FreeZBuffer();
      retval &= AllocateZBuffer();
    }

  _clip.Set(0,0,_width,_height);

  RestoreColortable();

  return retval;
}


////
////
//
// Colormap utilities
//
////
////

void SVGAGraphicPort::RestoreColortable(void)
{
  int c;

  for(c=0; c<GP_COLORMAP_SZ; c++)
    if(_colortable[c].Stat().Get(CC_USED))
      {
	ColorComponent r,g,b;
	_colortable[c].Get(&r,&g,&b);
	MapColor(c,r,g,b);
      }
}

void SVGAGraphicPort::ResetColortable(void)
{
  int c;

  for(c=0; c<GP_COLORMAP_SZ; c++)
    _colortable[c].Stat().Reset(CC_USED);
}

////
////
//
// High level functions
//
////
////

void SVGAGraphicPort::MapColor(const ColorIndex idx, 
			    const ColorComponent r, 
			    const ColorComponent g, 
			    const ColorComponent b )
{

  if(_bits_per_pixel > 8)
    {
      _colortable[idx].Stat().Set(CC_USED);
      _colortable[idx].Set(r,g,b);
      switch(_bits_per_pixel)
	{
	case 15:
	  _truecolormap[idx] = (((PixelValue)r >> 3) << 10) | 
	                       (((PixelValue)g >> 3) << 5)  | 
				((PixelValue)b >> 3);
	  break;
	case 24:
	case 32:
	  _truecolormap[idx] = (((PixelValue)r) << 16) | 
	                       (((PixelValue)g) << 8)  | 
				((PixelValue)b);
	  break;
	default:
	  (*this)[MSG_ERROR] << _bits_per_pixel << "bpp mode is not supported\n";
	  return;
	  break;
	}
      return;
    }


  _colortable[idx].Set(r,g,b);
  _colortable[idx].Stat().Set(CC_USED);

  vga_setpalette(idx,r >> 2,g >> 2,b >> 2);
}  

void SVGAGraphicPort::RGBMode(void)
{
  int r,g,b;

  (*this)[MSG_INFO] << "Switching to RGB mode\n";

  ResetColortable();

  for(r=0; r<4; r++)
     for(g=0; g<4; g++)
        for(b=0; b<4; b++)
	  MapColor(b + (g << 2) + (r << 4), r << 6, g << 6, b << 6);

  Attributes().Set(GPA_RGB);
  Attributes().Reset(GPA_CMAP);
}

void SVGAGraphicPort::ColormapMode(void)
{

  (*this)[MSG_INFO] << "Switching to colormap mode\n";

  ResetColortable();

// default colors

  MapColor(BLACK,   0,   0,   0   );
  MapColor(RED,     255, 0,   0   );
  MapColor(GREEN,   0,   255, 0   );
  MapColor(YELLOW,  255, 255, 0   );
  MapColor(BLUE,    0,   0,   255 );
  MapColor(MAGENTA, 255, 0,   255 );
  MapColor(CYAN,    0,   255, 255 );
  MapColor(WHITE,   255, 255, 255 );

  Attributes().Set(GPA_CMAP);
  Attributes().Reset(GPA_RGB);
}


int SVGAGraphicPort::SingleBuffer(void)
{
  if(!_linear)
    {
      (*this)[MSG_WARNING] << "SingleBuffer mode not available for non-linear framebuffers\n";
      return 0;
    }
  _graph_mem = _r_graph_mem;
  if(Attributes().Get(GPA_DBUFF))
    FreeImgBuffer();
  Attributes().Reset(GPA_DBUFF);
  (*this)[MSG_INFO] << "switching to SingleBuffer mode\n";
  return 1;
}

int SVGAGraphicPort::DoubleBuffer(void)
{
  if(!Attributes().Get(GPA_DBUFF))
    {
      if(!AllocateImgBuffer())
	return 0;
    }
  Attributes().Set(GPA_DBUFF);
  _graph_mem = _v_graph_mem;
  return 1;
}

int SVGAGraphicPort::SwapBuffers(void)
{
  if(Attributes().Get(GPA_DBUFF))
    {
//      vga_waitretrace();

      if(_linear)
	memcpy(_r_graph_mem, _v_graph_mem, _size);
      else
	{
	  int i;
	  char *graph_ptr = _v_graph_mem;
	  
	  for(i=0; i<_pages; i++)
	    {
	      vga_setpage(i);
	      memcpy(_r_graph_mem, graph_ptr, 65536);
	      graph_ptr += 65536;
	    }
	  if(_pad)
	    {
	      vga_setpage(_pages);
	      memcpy(_r_graph_mem, graph_ptr, _pad);
	    }
	}
    }
  return 1;
}


int SVGAGraphicPort::WaitEvent(void)
{
  return 1; // not implemented yet ! 
            // will try with a select() call on stdin and mouse fildes ... 
}

char SVGAGraphicPort::GetKey(void)
{
  return vga_getkey();
}

int  SVGAGraphicPort::GetMouse(ScrCoord *x, ScrCoord *y)
{
  int buttons, retval = 0;
  mouse_update();
  *x = mouse_getx();
  *y = mouse_gety();


// buttons are in reverse order than X
  buttons = mouse_getbutton();
  if(buttons & 1)
    retval |= 4;
  if(buttons & 2)
    retval |= 2;
  if(buttons & 4)
    retval |= 1;

  return retval;
}

int SVGAGraphicPort::ZBuffer(const int yes)
{
  if(yes && !(Attributes().Get(GPA_ZBUFF)))
    {
      AllocateZBuffer();
      Attributes().Set(GPA_ZBUFF);
    }
  else
    if(Attributes().Get(GPA_ZBUFF))
      {
	FreeZBuffer();
	Attributes().Reset(GPA_ZBUFF);
      }
}
  

int SVGAGraphicPort::SetGeometry(const ScrCoord width, const ScrCoord height)
{
  return Resize(width, height);
}


GraphicPort *MakeSVGAGraphicPort(char *name, 
				 ScrCoord width, 
				 ScrCoord height, 
				 int verb)
{

  cerr << "SVGA driver" << endl;

  SVGAGraphicPort *GP = new SVGAGraphicPort(name, width, height, verb);
  if(GP->ErrorCode())
    {
      delete GP;
      return NULL;
    }
  return GP;
}

class SVGAGraphicPortStub
{
public:
  SVGAGraphicPortStub(void) 
    { 
      GraphicPort_VC.Register(MakeSVGAGraphicPort,GP_VC_HANGS); 
    } 
};

static SVGAGraphicPortStub Dummy;
